/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

using namespace fe;

void HobbleOp::initialize(void)
{
	catalog<Real>("Envelope")=1.0;
	catalog<Real>("Envelope","min")= -100.0;
	catalog<Real>("Envelope","low")=0.0;
	catalog<Real>("Envelope","high")=1.0;
	catalog<Real>("Envelope","max")=100.0;
	catalog<String>("Envelope","page")="Restraints";
	catalog<String>("Envelope","hint")="Scale the overall effect.";

	catalog<bool>("TranslateLock")=true;
	catalog<String>("TranslateLock","label")="Translation Lock";
	catalog<String>("TranslateLock","page")="Restraints";
	catalog<bool>("TranslateLock","joined")=true;
	catalog<String>("TranslateLock","hint")="Prevent input from shifting.";

	catalog<Real>("TranslateEnv")=1.0;
	catalog<String>("TranslateEnv","label")="Translate";
	catalog<Real>("TranslateEnv","min")= -100.0;
	catalog<Real>("TranslateEnv","low")=0.0;
	catalog<Real>("TranslateEnv","high")=1.0;
	catalog<Real>("TranslateEnv","max")=100.0;
	catalog<String>("TranslateEnv","enabler")="TranslateLock";
	catalog<String>("TranslateEnv","page")="Restraints";
	catalog<String>("TranslateEnv","hint")="Scale just the translation.";

	catalog<bool>("RotateLock")=true;
	catalog<String>("RotateLock","label")="Rotation Lock";
	catalog<String>("RotateLock","page")="Restraints";
	catalog<bool>("RotateLock","joined")=true;
	catalog<String>("RotateLock","hint")="Prevent input from spinning.";

	catalog<Real>("RotateEnv")=1.0;
	catalog<String>("RotateEnv","label")="Rotate";
	catalog<Real>("RotateEnv","min")= -100.0;
	catalog<Real>("RotateEnv","low")=0.0;
	catalog<Real>("RotateEnv","high")=1.0;
	catalog<Real>("RotateEnv","max")=100.0;
	catalog<String>("RotateEnv","enabler")="RotateLock";
	catalog<String>("RotateEnv","page")="Restraints";
	catalog<String>("RotateEnv","hint")="Scale just the translation.";

	catalog<bool>("GroupedX")=false;
	catalog<String>("GroupedX","label")="Use Forward Group";
	catalog<bool>("GroupedX","joined")=true;
	catalog<String>("GroupedX","page")="Restraints";
	catalog<String>("GroupedX","hint")=
			"Use points on reference mesh to indicate forward tipping.";

	catalog<String>("GroupX");
	catalog<String>("GroupX","label")="Forward Point Group";
	catalog<String>("GroupX","suggest")="pointGroups";
	catalog<String>("GroupX","enabler")="GroupedX";
	catalog<String>("GroupX","page")="Restraints";
	catalog<String>("GroupX","hint")=
			"Points on reference mesh that indicate forward tipping"
			" relative to the center.";

	catalog<bool>("GroupedZ")=false;
	catalog<String>("GroupedZ","label")="Use Side Group";
	catalog<bool>("GroupedZ","joined")=true;
	catalog<String>("GroupedZ","page")="Restraints";
	catalog<String>("GroupedZ","hint")=
			"Use points on reference mesh to indicate sideward tipping.";

	catalog<String>("GroupZ");
	catalog<String>("GroupZ","label")="Side Point Group";
	catalog<String>("GroupZ","suggest")="pointGroups";
	catalog<String>("GroupZ","enabler")="GroupedZ";
	catalog<String>("GroupZ","page")="Restraints";
	catalog<String>("GroupZ","hint")=
			"Points on reference mesh that indicate sideward tipping"
			" relative to the center.";

	catalog<bool>("GroupedT")=false;
	catalog<String>("GroupedT","label")="Use Center Group";
	catalog<bool>("GroupedT","joined")=true;
	catalog<String>("GroupedT","page")="Restraints";
	catalog<String>("GroupedT","hint")=
			"Use points on reference mesh to calculate center.";

	catalog<String>("GroupT");
	catalog<String>("GroupT","label")="Center Point Group";
	catalog<String>("GroupT","suggest")="pointGroups";
	catalog<String>("GroupT","enabler")="GroupedT";
	catalog<String>("GroupT","page")="Restraints";
	catalog<String>("GroupT","hint")=
			"Points on reference mesh whose average indicates"
			" the central pivot";

	catalog<bool>("ForceArms")=false;
	catalog<String>("ForceArms","label")="Force Indexes";
	catalog<String>("ForceArms","page")="Restraints";
	catalog<bool>("ForceArms","joined")=true;
	catalog<String>("ForceArms","hint")=
			"Explicitly choose rotation control points.";

	catalog<I32>("ArmIndexX")= -1;
	catalog<I32>("ArmIndexX","min")= -1;
	catalog<I32>("ArmIndexX","max")=INT_MAX;
	catalog<String>("ArmIndexX","label")="Index To Side";
	catalog<String>("ArmIndexX","page")="Restraints";
	catalog<bool>("ArmIndexX","joined")=true;
	catalog<String>("ArmIndexX","enabler")="ForceArms";
	catalog<String>("ArmIndexX","hint")=
		"Explicit point index towards one side of object,"
		" presumably well behaved (-1 to automatically choose).";

	catalog<I32>("ArmIndexZ")= -1;
	catalog<I32>("ArmIndexZ","min")= -1;
	catalog<I32>("ArmIndexZ","max")=INT_MAX;
	catalog<String>("ArmIndexZ","label")="Index To Front";
	catalog<String>("ArmIndexZ","page")="Restraints";
	catalog<String>("ArmIndexZ","enabler")="ForceArms";
	catalog<String>("ArmIndexZ","hint")=
		"Explicit point index towards front of object,"
		" presumably well behaved (-1 to automatically choose).";

	catalog<bool>("Force Source")=false;
	catalog<String>("Force Source","page")="Restraints";
	catalog<bool>("Force Source","joined")=true;
	catalog<String>("Force Source","hint")="Explicitly choose input center.";

	catalog<SpatialVector>("Source")=SpatialVector(0.0,0.0,0.0);
	catalog<String>("Source","page")="Restraints";
	catalog<String>("Source","enabler")="Force Source";
	catalog<String>("Source","hint")="Explicit center of input.";

	catalog<bool>("Force Destination")=false;
	catalog<String>("Force Destination","page")="Restraints";
	catalog<bool>("Force Destination","joined")=true;
	catalog<String>("Force Destination","hint")="Explicitly place result.";

	catalog<SpatialVector>("Destination")=SpatialVector(0.0,0.0,0.0);
	catalog<String>("Destination","page")="Restraints";
	catalog<String>("Destination","enabler")="Force Destination";
	catalog<String>("Destination","hint")="Explicit placement of result.";

	catalog<SpatialVector>("Input Center");
	catalog<String>("Input Center","IO")="output";
	catalog<String>("Input Center","page")="Output";
	catalog<String>("Input Center","hint")=
			"Approximate center of input surface.";

	catalog<SpatialVector>("Reference Center");
	catalog<String>("Reference Center","IO")="output";
	catalog<String>("Reference Center","page")="Output";
	catalog<String>("Reference Center","hint")=
			"Approximate center of reference surface.";

	catalog<SpatialVector>("ConversionX");
	catalog<String>("ConversionX","label")="Rotation X";
	catalog<String>("ConversionX","page")="Output";
	catalog<String>("ConversionX","IO")="output";

	catalog<SpatialVector>("ConversionY");
	catalog<String>("ConversionY","label")="Rotation Y";
	catalog<String>("ConversionY","page")="Output";
	catalog<String>("ConversionY","IO")="output";

	catalog<SpatialVector>("ConversionZ");
	catalog<String>("ConversionZ","label")="Rotation Z";
	catalog<String>("ConversionZ","page")="Output";
	catalog<String>("ConversionZ","IO")="output";

	catalog<SpatialVector>("ConversionT");
	catalog<String>("ConversionT","label")="Translation";
	catalog<String>("ConversionT","page")="Output";
	catalog<String>("ConversionT","IO")="output";

	catalog<SpatialVector>("ConversionE");
	catalog<String>("ConversionE","label")="Euler XYZ";
	catalog<String>("ConversionE","page")="Output";
	catalog<String>("ConversionE","IO")="output";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Reference Surface");
	catalog<bool>("Reference Surface","optional")=true;
}

void HobbleOp::handle(Record& a_rSignal)
{
	SpatialVector& rInputCenter=catalog<SpatialVector>("Input Center");
	SpatialVector& rReferenceCenter=catalog<SpatialVector>("Reference Center");

	SpatialVector& rConversionX=catalog<SpatialVector>("ConversionX");
	SpatialVector& rConversionY=catalog<SpatialVector>("ConversionY");
	SpatialVector& rConversionZ=catalog<SpatialVector>("ConversionZ");
	SpatialVector& rConversionT=catalog<SpatialVector>("ConversionT");
	SpatialVector& rConversionE=catalog<SpatialVector>("ConversionE");

	const String groupX=catalog<String>("GroupX");
	const String groupZ=catalog<String>("GroupZ");
	const String groupT=catalog<String>("GroupT");
	const BWORD groupedX=(catalog<bool>("GroupedX") && !groupX.empty());
	const BWORD groupedZ=(catalog<bool>("GroupedZ") && !groupZ.empty());
	const BWORD groupedT=(catalog<bool>("GroupedT") && !groupT.empty());

	const Real envelope=catalog<Real>("Envelope");
	const bool forceArms=catalog<bool>("ForceArms");
	const I32 armIndexX=catalog<I32>("ArmIndexX");
	const I32 armIndexZ=catalog<I32>("ArmIndexZ");
	const bool forceSource=catalog<bool>("Force Source");
	const bool forceDestination=catalog<bool>("Force Destination");
	const SpatialVector source=catalog<SpatialVector>("Source");
	const SpatialVector destination=catalog<SpatialVector>("Destination");

	set(rInputCenter);
	set(rReferenceCenter);
	set(rConversionX,1.0,0.0,0.0);
	set(rConversionY,0.0,1.0,0.0);
	set(rConversionZ,0.0,0.0,1.0);
	set(rConversionT);

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,spOutputAccessible,e_point,e_position)) return;

	const U32 outputCount=spOutputPoint->count();
	if(!outputCount)
	{
		feLog("HobbleOp::handle point input count is zero\n");
		catalog<String>("error")="input points required";
		return;
	}

	sp<SurfaceAccessibleI> spReferenceAccessible;
	access(spReferenceAccessible,"Reference Surface",e_quiet);

	sp<SurfaceAccessorI> spReferencePoint;
	if(spReferenceAccessible.isValid())
	{
		access(spReferencePoint,spReferenceAccessible,
				e_point,e_position,e_quiet);
	}

	U32 referenceCount=0;
	if(spReferencePoint.isValid())
	{
		referenceCount=spReferencePoint->count();
		if(referenceCount!=outputCount)
		{
			catalog<String>("warning")+=
					"reference point count different than input point count; ";
		}
	}

	const BWORD translationLock=catalog<bool>("TranslateLock");
	const Real translateEnv=catalog<Real>("TranslateEnv");

	BWORD rotationLock=spReferencePoint.isValid() &&
			catalog<bool>("RotateLock");
	const Real rotateEnv=catalog<Real>("RotateEnv");

	if(!translationLock && !rotationLock)
	{
		return;
	}
	if(translateEnv==0.0 && rotateEnv==0.0)
	{
		return;
	}

	Real indexX=0;
	Real indexZ=0;

	if(!forceArms || !forceSource || !forceDestination)
	{
		const U32 start=(forceArms && forceSource);
		const U32 passes=1+(spReferencePoint.isValid() &&
				(!forceArms || !forceDestination));
		for(U32 pass=start;pass<passes;pass++)
		{
			sp<SurfaceAccessorI> spPoint=pass? spReferencePoint: spOutputPoint;
			FEASSERT(spPoint.isValid());

			SpatialVector min=spPoint->spatialVector(0);
			SpatialVector max=min;
			SpatialVector sum=min;
			for(U32 index=1;index<outputCount;index++)
			{
				const SpatialVector point=spPoint->spatialVector(index);
				sum+=point;

				for(U32 m=0;m<3;m++)
				{
					const Real value=point[m];
					if(min[m]>value)
					{
						min[m]=value;
					}
					if(max[m]<value)
					{
						max[m]=value;
						if(m==0)
						{
							indexX=index;
						}
						else if(m==2)
						{
							indexZ=index;
						}
					}
				}
			}

			if(pass)
			{
//				rReferenceCenter=0.5*(min+max);
				rReferenceCenter=sum/Real(outputCount);
			}
			else
			{
//				rInputCenter=0.5*(min+max);
				rInputCenter=sum/Real(outputCount);
			}
		}
	}

	SpatialVector referenceTrue=rReferenceCenter;

	if(forceSource)
	{
		rInputCenter=source;
	}
	if(forceDestination)
	{
		rReferenceCenter=destination;
	}
	if(forceArms)
	{
		if(armIndexX>=0)
		{
			indexX=armIndexX;
		}
		if(armIndexZ>=0)
		{
			indexZ=armIndexZ;
		}
	}
	if(rotationLock)
	{
		if(indexX==indexZ)
		{
			catalog<String>("warning")+=
					"rotation control indexes can not match"
					" (rotation lock dismissed); ";
			rotationLock=FALSE;
		}
		if(indexX>=outputCount || indexZ>=outputCount)
		{
			catalog<String>("warning")+=
					"a rotation control index exceeds input point count"
					" (rotation lock dismissed); ";
			rotationLock=FALSE;
		}
		if(indexX>=referenceCount || indexZ>=referenceCount)
		{
			catalog<String>("warning")+=
					"a rotation control index exceeds reference point count"
					" (rotation lock dismissed); ";
			rotationLock=FALSE;
		}
	}

	const SpatialVector zero(0.0,0.0,0.0);
	SpatialTransform conversion;
	if(rotationLock)
	{
		//* rotation (translation optional)

		SpatialTransform xformRef;
		SpatialTransform xformDef;

		for(U32 pass=0;pass<2;pass++)
		{
			sp<SurfaceAccessorI> spPoint=pass? spReferencePoint: spOutputPoint;
			sp<SurfaceAccessorI> spGroup;

			SpatialVector center=pass? referenceTrue: rInputCenter;

			U32 count=0;

			if(groupedT && (pass? !forceDestination: !forceSource) &&
					access(spGroup,spReferenceAccessible,e_pointGroup,groupT)
					&& (count=spGroup->count()))
			{
				set(center);
				for(U32 index=0;index<count;index++)
				{
					const I32 pointIndex=spGroup->integer(index);
					FEASSERT(pointIndex>=0);
					center+=spPoint->spatialVector(pointIndex);
				}
				center*=1.0/Real(count);

				if(pass)
				{
					rReferenceCenter=center;
					referenceTrue=center;
				}
				else
				{
					rInputCenter=center;
				}
			}

			SpatialVector side;
			if(groupedX &&
					access(spGroup,spReferenceAccessible,e_pointGroup,groupX)
					&& (count=spGroup->count()))
			{
				set(side);
				for(U32 index=0;index<count;index++)
				{
					const U32 pointIndex=spGroup->integer(index);
					side+=spPoint->spatialVector(pointIndex);
				}
				side*=1.0/Real(count);
			}
			else
			{
				side=spPoint->spatialVector(indexX);
			}
			side-=center;

			SpatialVector forward;
			if(groupedZ &&
					access(spGroup,spReferenceAccessible,e_pointGroup,groupZ)
					&& (count=spGroup->count()))
			{
				set(forward);
				for(U32 index=0;index<count;index++)
				{
					const U32 pointIndex=spGroup->integer(index);
					forward+=spPoint->spatialVector(pointIndex);
				}
				forward*=1.0/Real(count);
			}
			else
			{
				forward=spPoint->spatialVector(indexZ);
			}
			forward-=center;

			BWORD success=makeFrame(pass? xformRef: xformDef,
					(pass && translationLock)? rReferenceCenter: rInputCenter,
					unitSafe(side),forward);
		}

		SpatialTransform invDef;
		invert(invDef,xformDef);
		conversion=invDef*xformRef;

		if(envelope*translateEnv!=1.0 || envelope*rotateEnv!=1.0)
		{
			Real power=1.0;
			Real envT=1.0;
			Real envR=1.0;

			if(fabs(rotateEnv)>fabs(translateEnv))
			{
				power=envelope*rotateEnv;
				envT=translateEnv/rotateEnv;
			}
			else
			{
				power=envelope*translateEnv;
				envR=rotateEnv/translateEnv;
			}

			if(envT!=1.0 || envR!=1.0)
			{
				const SpatialVector transMod=xformDef.translation()+
						envT*(xformRef.translation()-xformDef.translation());

				SpatialQuaternion quatRef(xformRef);
				SpatialQuaternion quatDef(xformDef);
				SpatialQuaternion quatMod;
				slerp(quatMod,quatDef,quatRef,envR);

				SpatialTransform xformMod=quatMod;

				setTranslation(xformMod,transMod);

				conversion=invDef*xformMod;
			}

			if(power!=1.0)
			{
				MatrixPower<SpatialTransform> matrixPower;
				matrixPower.solve(conversion,conversion,power);
			}
		}
	}
	else
	{
		//* translation only
		setIdentity(conversion);
		translate(conversion,
				envelope*translateEnv*(rReferenceCenter-rInputCenter));
	}

	rConversionX=conversion.column(0);
	rConversionY=conversion.column(1);
	rConversionZ=conversion.column(2);
	rConversionT=conversion.column(3);

	SpatialEuler eulerXYZ=conversion;
	rConversionE=eulerXYZ/degToRad;

	for(U32 index=0;index<outputCount;index++)
	{
		const SpatialVector point=spOutputPoint->spatialVector(index);
		SpatialVector transformed;
		transformVector(conversion,point,transformed);

		spOutputPoint->set(index,transformed);
	}
}
