/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "plugin.pmh"

namespace fe
{

Master::Master(void):
	m_active(false)
{
#if FE_CODEGEN<=FE_DEBUG
	printf("Creating FE Master %p\n", this);
#endif

	setName("Master");

#if FE_CODEGEN<FE_OPTIMIZE
	SystemTicker::calibrate();
#endif

#if FE_CHECK_NEW
	if(!Memory::newWorks())
	{
		feLog("Warning: not using FE version of new()\n");
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
		feLog("  heap unification must be provided by your alternative\n");
#elif FE_MEM_ALIGNMENT && FE_OS!=FE_OSX
		feLog("  memory alignment must be provided by your alternative\n");
#endif
	}
#endif

	m_spAllocator = new BasicAllocator();

	sp<TypeMaster> spTypeMaster(new TypeMaster());
	bind(spTypeMaster);

	sp<Registry> spRegistry(new Registry(this));
	bind(spRegistry);

	m_active=true;

	assertCore(typeMaster());
	assertPlugin(typeMaster());

#if FE_COUNTED_TRACK
	registerRegion(this,sizeof(Master));
#endif
#if FE_CODEGEN<=FE_DEBUG
	feLog("Master created\n");
#endif
}

Master::Master(const Master &other)
{
	if(&other != this)
	{
		feX("fe::Master::Master","clone impossible");
	}
}

Master::~Master(void)
{
#if FE_CODEGEN<=FE_DEBUG
	printf("Dismissing FE Master %p\n", this);
#endif

	if(m_spCatalog.isValid())
	{
#if FE_CODEGEN<=FE_DEBUG
		catalog()->catalogDump();
#endif
		catalog()->catalogClear();
#if FE_CODEGEN<=FE_DEBUG
		feLog("catalog cleared\n");
#endif
	}

	m_active=false;
	m_spTypeMaster = NULL;
	m_spCatalog=NULL;
	m_spAllocator = NULL;
#if FE_COUNTED_TRACK
	deregisterRegion(this);
#endif
	//* component tracker is tied to the registry
	m_spRegistry = NULL;
#if FE_CODEGEN<=FE_DEBUG
	feLog("~Master done\n");
#endif
}

sp<Catalog> Master::catalog(void)
{
	if(!m_active)
	{
		return sp<Catalog>();
	}
	if(!m_spCatalog.isValid())
	{
		if(!m_spLibrary.isValid())
		{
			m_spLibrary=new Library();
			m_spLibrary->setRegistry(m_spRegistry);
		}
		if(!m_spLibrary.isValid())
		{
			return sp<Catalog>();
		}

		m_spCatalog=Library::create<Catalog>(
				"Master Catalog",m_spLibrary.raw());

		if(!m_spCatalog->cataloged("path:plugin"))
		{
			String pluginPath=String("${FE_PLUGIN}").substituteEnv();
			if(pluginPath.empty())
			{
				pluginPath="${FE_ROOT}/../plugin";
			}
			m_spCatalog->catalog<String>("path:plugin")=
					pluginPath.substituteEnv();
		}
	}
	return m_spCatalog;
}

} /* namespace */
